import numpy as np
cimport numpy as np

from libc.math cimport exp as c_exp
from libc.math cimport sin as c_sin
from libc.math cimport cos as c_cos

from cython.parallel import prange, parallel
from cython import boundscheck, wraparound, nonecheck, cdivision
from cpython cimport array
import array

#ZUM PARALLELISIEREN MIT CYTHON
#1.  Inplace Operatoren ("+=", etc) kommen hier eine besondere Bedeutung zu.
#2.  Bei Cython-Version 0.25.2 gibt es einen Bug beim Zuweisen von numpy-Arrays:
#    Wenn ein Vektor zuerst vollgeschrieben und im späteren Verlauf der Methode 
#    wieder ausgelesen wird, wurden nicht die korrekten Werte ausgelesen.
#    Wurde der Vektor/Matrix allerdings nur einmalig geschrieben und dann returned,
#    war alles okay. Ist im Code unten nochmal am Bsp. "beamformerCython" beschrieben.
#    Dieses Problem tritt nur auf, falls der Code über prange parallelisiert wird.
#    Außerdem tritt der Fehler nicht immer auf. Liegt vllt an unausgereifter Ausnutzung
#    der gesharedten Caches.
#3.  Welche Variablen Thread-private und welche geshared werden sollen, entscheidet
#    Cython selbst.
#4.  Beim Beamformer (s.u.) hat das parallelisieren mit 'prange' zu erheblichen Einbußen in
#    der Laufzeit geführt (ca 5x LANGSAMER). Es gibt allerdings einfache BSPs
#    ("c_arrayOpenMP", "c_array", s.u.), bei denen mit gleichen Compiler-Optionen 
#    wie beim Beamformer durch die Parallelisierung mit 'prange' erhebliche 
#    Laufzeitverbesserungen erreicht werden.
#5.  Es muss ein setup-File mit den Kompilieroptionen erstellt werden. In diesem 
#    Fall ist dies "setupCythonOpenMP.py". Danach wird über die Konsole mit dem
#    Befehl "python setupCythonOpenMP.py build_ext --inplace" kompiliert.




#==============================================================================
# # Das Ganze wird fuer effizientes indexing gebraucht...
DTYPEComplex = np.complex128
DTYPEReal = np.float
ctypedef np.complex128_t DTYPEComplex_t
ctypedef np.float_t DTYPEReal_t
#==============================================================================

# Die Dekoratoren sollen nochmals Schnelligkeit bringen.
@nonecheck(False)
@boundscheck(False)
@wraparound(False)
@cdivision(True)
def beamformerCython(np.ndarray[DTYPEComplex_t, ndim=3] csm, np.ndarray[DTYPEReal_t] r0, np.ndarray[DTYPEReal_t, ndim=2] rm, np.ndarray[DTYPEComplex_t] kj):
#def beamformerCython(complex[:,:,:] csm, double[:] r0, double[:,:] rm, complex[:] kj):  # Standard Methodenkopf (laut Doku etwas langsamer)
    """ Dies ist eine Implementierung der Methode 'r_beamfull_inverse' der 
    mit scipy.weave erzeugten 'beamformer.so'. Benutzt Parallelisierung via 
    cythons 'prange' und Kompilierung mit OpenMP.
    """
    cdef int nFreqs = csm.shape[0]
    cdef int nGridPoints = len(r0)
    cdef int nMics = csm.shape[1]
    
    cdef double[:,:] beamformOutput = np.zeros([nFreqs, nGridPoints], dtype=DTYPEReal)
    cdef complex[:] steerVec = np.zeros([nMics], dtype=DTYPEComplex)

    cdef int cntFreqs, cntGrid, cntMics, cntMics2, cntMics1
    cdef double r01, kjj, rm1, temp1, rs, temp3
    cdef complex temp2
#    cdef float temp3  # Das hat erstmal nicht auf Anhieb geklappt, deshalb ist temp3 noch double
    
    for cntFreqs in range(nFreqs):
        kjj = kj[cntFreqs].imag
        
        
        
        
        
        with nogil, parallel():
        
            
            # Mit Chef drüber reden.
            
            
            
            
            for cntGrid in prange(nGridPoints, schedule='static'):  # Parallelisierung
                r01 = r0[cntGrid]
                rs = r01 ** 2
                temp1 = 0.0
                for cntFreqs in range(nFreqs):
                    kjj = kj[cntFreqs].imag
                    for cntGrid in range(nGridPoints):
                        rs = 0
                        r01 = r0[cntGrid]
                        for cntMics in range(nMics):
                            rm1 = rm[cntGrid, cntMics]
                            rs += 1.0 / (rm1 ** 2)
                            temp3 = (kjj * (rm1 - r01))
                            steerVec[cntMics] = (c_cos(temp3) - 1j * c_sin(temp3)) * rm1
                        rs = r01 ** 2
                        
                        temp1 = 0.0
                        for cntMics in range(nMics): 
                            temp2 = 0.0
                            for cntMics2 in range(cntMics):
                                temp2 = temp2 + csm[cntFreqs, cntMics2, cntMics] * steerVec[cntMics2]  # Falls nicht mit prange gearbeitet wird, kann man "+= etc" wie sonst benutzen
                            temp1 = temp1 + 2 * (temp2 * steerVec[cntMics].conjugate()).real
                            temp1 = temp1 + (csm[cntFreqs, cntMics, cntMics] * (steerVec[cntMics]).conjugate() * steerVec[cntMics]).real
                        beamformOutput[cntFreqs, cntGrid] = (temp1 / rs)        
    return beamformOutput

@boundscheck(False)
@wraparound(False)
@nonecheck(False)
@cdivision(True)
def beamformerCythonNOTparallel(np.ndarray[DTYPEComplex_t, ndim=3] csm, np.ndarray[DTYPEReal_t] r0, np.ndarray[DTYPEReal_t, ndim=2] rm, np.ndarray[DTYPEComplex_t] kj):
    """ Dies ist eine Implementierung der Methode 'r_beamfull_inverse' der 
    mit scipy.weave erzeugten 'beamformer.so', OHNE Parallelisierung.
    """
    cdef int nFreqs = csm.shape[0]
    cdef int nGridPoints = len(r0)
    cdef int nMics = csm.shape[1]
    
    cdef double[:,:] beamformOutput = np.zeros([nFreqs, nGridPoints], dtype=DTYPEReal)
    cdef complex[:] steerVec = np.zeros([nMics], dtype=DTYPEComplex)
    
    cdef int cntFreqs, cntGrid, cntMics, cntMics2
    cdef double rs, r01, kjj, rm1, temp1, temp3
    cdef complex temp2
#    cdef float temp3
    
    for cntFreqs in range(nFreqs):
        kjj = kj[cntFreqs].imag
        for cntGrid in range(nGridPoints):
            rs = 0
            r01 = r0[cntGrid]
            for cntMics in range(nMics):
                rm1 = rm[cntGrid, cntMics]
                rs += 1.0 / (rm1 ** 2)
                temp3 = (kjj * (rm1 - r01))
                steerVec[cntMics] = (c_cos(temp3) - 1j * c_sin(temp3)) * rm1
            rs = r01 ** 2
            
            temp1 = 0.0
            for cntMics in range(nMics): 
                temp2 = 0.0
                for cntMics2 in range(cntMics):
                    temp2 += csm[cntFreqs, cntMics2, cntMics] * steerVec[cntMics2]  # Falls nicht mit prange gearbeitet wird, kann man "+= etc" wie sonst benutzen
                temp1 += 2 * (temp2 * steerVec[cntMics].conjugate()).real
                temp1 += (csm[cntFreqs, cntMics, cntMics] * (steerVec[cntMics]).conjugate() * steerVec[cntMics]).real
            beamformOutput[cntFreqs, cntGrid] = (temp1 / rs)          
    return beamformOutput

@nonecheck(False)
@boundscheck(False)
@wraparound(False)
@cdivision(True)
def beamformerCythonCorrectButSlow(np.ndarray[DTYPEComplex_t, ndim=3] csm, np.ndarray[DTYPEReal_t] r0, np.ndarray[DTYPEReal_t, ndim=2] rm, np.ndarray[DTYPEComplex_t] kj):
#def beamformerCython(complex[:,:,:] csm, double[:] r0, double[:,:] rm, complex[:] kj):  # Standard Methodenkopf (laut Doku etwas langsamer)
    """ Dies ist eine Implementierung der Methode 'r_beamfull_inverse' der 
    mit scipy.weave erzeugten 'beamformer.so'. Benutzt Parallelisierung via 
    cythons 'prange' und Kompilierung mit OpenMP.
    """
    cdef int nFreqs = csm.shape[0]
    cdef int nGridPoints = len(r0)
    cdef int nMics = csm.shape[1]
    
    cdef double[:,:] beamformOutput = np.zeros([nFreqs, nGridPoints], dtype=DTYPEReal)

    cdef int cntFreqs, cntGrid, cntMics, cntMics2, cntMics1
    cdef double r01, kjj, rm1, temp1, rs, temp3
    cdef complex temp2, steerVec1, steerVec
#    cdef float temp3
    
    for cntFreqs in range(nFreqs):
        kjj = kj[cntFreqs].imag
        for cntGrid in prange(nGridPoints, nogil=True, schedule='static'):  # Parallelisierung
            r01 = r0[cntGrid]
            rs = r01 ** 2
            
            # Im ursprünglichen Code wurde hier erstmal in einer Schleife 
            # ueber die Mics der Steering-Vektor erstellt und später verarbeitet. 
            # Cython verhaut aber beim späteren Auslesen dieses Vektors, falls 
            # mit prange parallelisiert wird (siehe 2. im Kommentar ganz oben).
            # Deshalb wurde hier der Code ein wenig geändert (Wirkung bleibt aber erhalten, 
            # allerdings werdfen redundante Operationen ausgeführt -->langsamer)
            # Siehe Methode "beamformerCythonNOTparallel" für den ursprünglichen Code.
            
            temp1 = 0.0
            for cntMics in range(nMics): 
                temp2 = 0.0
                rm1 = rm[cntGrid, cntMics]
                temp3 = (kjj * (rm1 - r01))
                steerVec = (c_cos(temp3) - 1j * c_sin(temp3)) * rm1
                
                for cntMics2 in range(cntMics):
                    rm1 = rm[cntGrid, cntMics2]
                    temp3 = (kjj * (rm1 - r01))
                    steerVec1 = (c_cos(temp3) - 1j * c_sin(temp3)) * rm1
                    temp2 = temp2 + csm[cntFreqs, cntMics2, cntMics] * steerVec1
                temp1 = temp1 + 2 * (temp2 * steerVec.conjugate()).real
                temp1 = temp1 + (csm[cntFreqs, cntMics, cntMics] * steerVec.conjugate() * steerVec).real
            beamformOutput[cntFreqs, cntGrid] = (temp1 / rs)  # Hier gibt es keine Schwierigkeiten bzgl des Stichpkts 2. von oben. Nur falls "beamformOutput" nochmal in derselben Methode ausgelesen werden sollte.
    return beamformOutput


#==============================================================================
# BSP, FÜR DIE DIE PARALLELISIERUNG MIT PRANGE LAUFZEITVERBESSERUNGEN BRINGT:
#---------------------------------------------------------------------------
#
# Die nachfolgenden 2 Fkt. machen das gleiche. Allerdings ist "c_arrayOpenMP" mit 'prange' parallelisiert.
# Macht man von ausserhalb einen Laufzeitvergleich, ist zu erkennen, dass das
# Parallelisieren hier tatsächlich was bringt (bei mir ungefähr 2x schneller).
@boundscheck(False)
@wraparound(False)
@nonecheck(False)
def c_array(double[:] X):
    cdef int N = X.shape[0]
    cdef double[:] Y = np.zeros(N)
    cdef int i
    for i in range(N):  # NICHT parallelisiert
        if X[i] > 0.5:
            Y[i] = c_exp(X[i])
        else:
            Y[i] = 0
    return Y

@boundscheck(False)
@wraparound(False)
@nonecheck(False)
def c_arrayOpenMP(double[:] X):
    cdef int N = X.shape[0]
    cdef double[:] Y = np.zeros(N)
    cdef int i
    for i in prange(N, nogil=True):  # parallelisiert
        if X[i] > 0.5:
            Y[i] = c_exp(X[i])
        else:
            Y[i] = 0
    return Y



# -*- coding: utf-8 -*-
#pylint: disable-msg=E0611, E1101, C0103, R0901, R0902, R0903, R0904, W0232
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------
"""
Implements tools for Acoular.

.. autosummary::
    :toctree: generated/
    
    MetricEvaluator
    return_result
    spherical_hn1
    get_radiation_angles
    get_modes
    barspectrum
    bardata
"""

import acoular as ac
from numpy import array, concatenate, newaxis, where,arctan2,sqrt,pi,mod,zeros,\
    complex128, ones, inf, minimum, empty
from numpy.linalg import norm
from numpy.ma import masked_where
from .spectra import synthetic
from traits.api import Bool, CArray, Float, HasPrivateTraits, Instance, Property, Either

from scipy.special import spherical_yn, spherical_jn, sph_harm
from scipy.spatial.distance import cdist


class MetricEvaluator(HasPrivateTraits):
    """Evaluate the reconstruction performance of source mapping methods.

    This class can be used to calculate the following performance metrics
    according to Herold and Sarradj (2017):
    * Specific level error
    * Overall level error
    * Inverse level error
    """

    #: an array of shape=(nf,ng) containing the squared sound pressure data of the   
    #: source mapping. (nf: number of frequencies, ng: number of grid points)
    data = CArray(shape=(None,None),
        desc="Contains the calculated squared sound pressure values in Pa**2.")

    #: an array of shape=(nf,ns) containing the squared sound pressure data of the
    #: ground-truth sources. (nf: number of frequencies, ns: number of sources)
    target_data = CArray(shape=(None,None),
        desc="Contains the ground-truth squared sound pressure values in Pa**2.")

    #: :class:`~acoular.grids.Grid`-derived object that provides the grid locations
    #: for the calculated source mapping data.
    grid = Instance(ac.Grid,
        desc="Grid instance that belongs to the calculated data")

    #: :class:`~acoular.grids.Grid`-derived object that provides the grid locations
    #: for the ground-truth data.
    target_grid = Instance(ac.Grid,
        desc="Grid instance that belongs to the ground-truth data")

    #: sector type. Currently only circular sectors are supported.
    sector = Instance(ac.CircSector, default=ac.CircSector(),)

    #: if set True: use shrink integration area if two sources are closer
    #: than 2*r. The radius of the integration area is then set to half the
    #: distance between the two sources.
    adaptive_sector_size = Bool(True,
        desc="adaptive integration area")

    #: if set `True`, the same amplitude can be assigned to multiple targets if
    #: the integration area overlaps. If set `False`, the amplitude is assigned
    #: to the first target and the other targets are ignored.
    multi_assignment = Bool(True,
        desc="if set True, the same amplitude can be assigned to multiple targets if the integration area overlaps")

    #: returns the determined sector sizes for each ground-truth source position
    sectors = Property()

    def _validate_shapes(self):
        if self.data.shape[0] != self.target_data.shape[0]:
            raise ValueError("data and target_data must have the same number of frequencies!")
        if self.data.shape[1] != self.grid.size:
            raise ValueError("data and grid must have the same number of grid points!")
        if self.target_data.shape[1] != self.target_grid.size:
            raise ValueError("target_data and target_grid must have the same number of grid points!")

    def _get_sector_radii(self):
        ns = self.target_data.shape[1]
        radii = ones(ns)*self.sector.r
        if self.adaptive_sector_size:
            locs = self.target_grid.gpos.T
            intersrcdist = cdist(locs, locs)
            intersrcdist[intersrcdist == 0] = inf
            intersrcdist = intersrcdist.min(0)/2
            radii = minimum(radii,intersrcdist)
        return radii

    def _get_sectors(self):
        """Returns a list of CircSector objects for each target location."""
        r = self._get_sector_radii()
        ns = self.target_data.shape[1]
        sectors = []
        for i in range(ns):
            loc = self.target_grid.gpos[:,i]
            sectors.append(
                ac.CircSector(r=r[i],x=loc[0],y=loc[1]))
        return sectors

    def _integrate_sectors(self):
        """Integrates over target sectors.

        Returns
        -------
        array (num_freqs,num_sources)
            returns the integrated Pa**2 values for each sector
        """
        results = empty(shape=self.target_data.shape)
        for f in range(self.target_data.shape[0]):
            data = self.data[f]
            for i in range(self.target_data.shape[1]):
                sector = self.sectors[i]
                results[f,i] = ac.integrate(data,self.grid,sector)
                if not self.multi_assignment:
                    indices = self.grid.subdomain(sector)
                    data[indices] = 0 # set values to zero (can not be assigned again)
        return results

    def get_overall_level_error(self):
        """Returns the overall level error (Herold and Sarradj, 2017).

        Returns
        -------
        numpy.array
            overall level error of shape=(nf,)
        """
        self._validate_shapes()
        return ac.L_p(self.data.sum(axis=1)) - ac.L_p(self.target_data.sum(axis=1))

    def get_specific_level_error(self):
        """Returns the specific level error (Herold and Sarradj, 2017).

        Returns
        -------
        numpy.array
            specific level error of shape=(nf,ns). nf: number of frequencies, ns: number of sources
        """
        self._validate_shapes()
        sector_result = self._integrate_sectors()
        return ac.L_p(sector_result) - ac.L_p(self.target_data)

    def get_inverse_level_error(self):
        """Returns the inverse level error (Herold and Sarradj, 2017).

        Returns
        -------
        numpy.array
            inverse level error of shape=(nf,1)
        """
        self._validate_shapes()
        sector_result = self._integrate_sectors()
        return ac.L_p(sector_result.sum(axis=1)) - ac.L_p(self.data.sum(axis=1))



def return_result(source, nmax=-1, num=128):
    """
    Collects the output from a 
    :meth:`SamplesGenerator.result()<acoular.tprocess.SamplesGenerator.result>`
    generator and returns an assembled array with all the data.
   
    Parameters
    ----------
    source: SamplesGenerator or derived object.
        This is the  :class:`SamplesGenerator<acoular.tprocess.SamplesGenerator>` data source.
    nmax: integer
        With this parameter, a maximum number of output samples can be set 
        (first dimension of array). If set to -1 (default), samples are 
        collected as long as the generator yields them.
    num : integer
        This parameter defines the size of the blocks that are fetched.
        Defaults to 128.
          
    Returns
    -------
    array of floats (number of samples, source.numchannels)
        Array that holds all the data.
    """
    resulter = (_.copy() for _ in source.result(num))
    
    if nmax > 0: 
        nblocks = (nmax-1) // num + 1
        return concatenate( 
                      list( res for _, res in 
                            zip(range(nblocks), 
                                resulter) ) )[:nmax]
    else:
        return concatenate(list(resulter))


def spherical_hn1(n,z,derivativearccos=False):
   """ Spherical Hankel Function of the First Kind 
   
   """
   return spherical_jn(n,z,derivative=False)+1j*spherical_yn(n,z,derivative=False) 

def get_radiation_angles(direction,mpos, sourceposition):
    """
    Returns azimuthal and elevation angles between the mics and the source 
    
    Parameters
    ----------
    direction : array of floats
        Spherical Harmonic orientation
    mpos : array of floats
        x, y, z position of microphones
    sourceposition : array of floats
        position of the source        
        
    Returns
    -------
    azi, ele : array of floats
        the angle between the mics and the source 
    """
    #direction of the Spherical Harmonics
    direc = array(direction, dtype = float)
    direc = direc/norm(direc)
    # distances
    source_to_mic_vecs = mpos-array(
        sourceposition).reshape((3, 1))
    source_to_mic_vecs[2] *= -1 # invert z-axis (acoular)    #-1
    # z-axis (acoular) -> y-axis (spherical)
    # y-axis (acoular) -> z-axis (spherical)
    #theta
    ele = arctan2(sqrt(source_to_mic_vecs[0]**2 + source_to_mic_vecs[2]**2),source_to_mic_vecs[1])
    ele +=arctan2(sqrt(direc[0]**2 + direc[2]**2), direc[1])  
    ele += pi*.5 # convert from [-pi/2, pi/2] to [0,pi] range
    #phi
    azi = arctan2(source_to_mic_vecs[2],source_to_mic_vecs[0]) 
    azi += arctan2(direc[2],direc[0]) 
    azi = mod(azi,2*pi)
    return azi, ele

def get_modes(lOrder, direction, mpos , sourceposition = array([0,0,0])):
    """
    Returns Spherical Harmonic Radiation Pattern at the Microphones
      
    Parameters
    ----------
    lOrder : int
        Maximal order of spherical harmonic
    direction : array of floats
        Spherical Harmonic orientation
    mpos : array of floats
        x, y, z position of microphones
    sourceposition : array of floats
        position of the source

    Returns
    -------
    modes : array of floats
        the radiation values at each microphone for each mode
    """
    azi, ele = get_radiation_angles(direction,mpos,sourceposition) # angles between source and mics 
    modes = zeros((azi.shape[0], (lOrder+1)**2), dtype=complex128)
    i = 0
    for l in range(lOrder+1):
        for m in range(-l, l+1):
            modes[:, i] = sph_harm(m, l, azi, ele)
            if m<0:
                modes[:, i]=modes[:, i].conj()*1j             
            i += 1
    return modes

def barspectrum(data, fftfreqs, num = 3, bar = True, xoffset = 0.0):
    """
    Returns synthesized frequency band values of spectral data to be plotted
    as bar graph with the matlpotlib plot command.
    
   
    Parameters
    ----------
    data : array of floats
        The spectral data (sound pressures in Pa) in an array with one value 
        per frequency line.
    fftfreqs : array of floats
        Discrete frequencies from FFT.
    num : integer
        Controls the width of the frequency bands considered; defaults to
        3 (third-octave band).
    bar : bool
        If True, returns bar-like curve. If False, normal plot (direct 
        line between data points) is returned.
    xoffset : float
        If bar is True, offset of the perpendicular line (helpful if 
        plotting several curves above each other). 
        
        ===  =====================
        num  frequency band width
        ===  =====================
        1    octave band
        3    third-octave band
        ===  =====================

    Returns
    -------
    (flulist, plist, fc)
    flulist : array of floats
        Lower/upper band frequencies in plottable format.
    plist : array of floats
        Corresponding synthesized frequency band values in plottable format.
    fc : array of floats
        Evaluated band center frequencies.
    """
    
    if num not in [1,3]:
        print('Only octave and third-octave bands supported at this moment.')
        return (0,0,0)
        

    # preferred center freqs after din en iso 266 for third-octave bands
    fcbase = array([31.5,40,50,63,80,100,125,160,200,250])
    # DIN band center frequencies from 31.5 Hz to 25 kHz
    fc = concatenate((fcbase, fcbase*10., fcbase[:]*100.))[::(3//num)]
    
    
    # exponent for band width calculation
    ep = 1. / (2.*num)
    
    # lowest and highest possible center frequencies 
    # for chosen band and sampling frequency
    f_low = fftfreqs[1]*2**ep
    f_high = fftfreqs[-1]*2**-ep
    # get possible index range
    if fc[0] >= f_low:
        i_low = 0
    else:
        i_low = where(fc < f_low)[0][-1]
    
    if fc[-1] <= f_high:
        i_high = fc.shape[0]
    else:
        i_high = where(fc > f_high)[0][0]
        
    # synthesize sound pressure values
    p = array([ synthetic(data, fftfreqs, list(fc[i_low:i_high]), num) ])
 
    if bar:
        # upper and lower band borders
        flu = concatenate(( fc[i_low:i_low+1]*2**-ep,
                            ( fc[i_low:i_high-1]*2**ep + fc[i_low+1:i_high]*2**-ep ) / 2.,
                            fc[i_high-1:i_high]*2**ep ))
        # band borders as coordinates for bar plotting
        flulist = 2**(2*xoffset*ep) * (array([1,1])[:,newaxis]*flu[newaxis,:]).T.reshape(-1)[1:-1]   
        # sound pressures as list for bar plotting
        plist = (array([1,1])[:,newaxis]*p[newaxis,:]).T.reshape(-1)
    else:
        flulist = fc[i_low:i_high]
        plist = p[0,:]
    #print(flulist.shape, plist.shape)
    return (flulist, plist, fc[i_low:i_high])
    


def bardata(data, fc, num=3, bar = True, xoffset = 0.0, masked = -360):
    """
    Returns data to be plotted
    as bar graph with the matlpotlib plot command.
    
   
    Parameters
    ----------
    data : array of floats
        The spectral data 
    fc : array of floats
        Band center frequencies
    bar : bool
        If True, returns bar-like curve. If False, normal plot (direct 
        line between data points) is returned.
    xoffset : float
        If bar is True, offset of the perpendicular line (helpful if 
        plotting several curves above each other). 
        
        ===  =====================
        num  frequency band width
        ===  =====================
        1    octave band
        3    third-octave band
        ===  =====================

    Returns
    -------
    (flulist, plist)
    flulist : array of floats
        Lower/upper band frequencies in plottable format.
    plist : array of floats
        Corresponding values in plottable format.
    """
    ep = 1. / (2.*num)
    
    if bar:
        # upper and lower band borders
        flu = concatenate(( fc[:1]*2**-ep,
                            ( fc[:-1]*2**ep + fc[1:]*2**-ep ) / 2.,
                            fc[-1:]*2**ep ))
        # band borders as coordinates for bar plotting
        flulist = 2**(xoffset*1./num) * (array([1,1])[:,newaxis]*flu[newaxis,:]).T.reshape(-1)[1:-1]   
        # sound pressures as list for bar plotting
        plist = (array([1,1])[:,newaxis] * data[newaxis,:]).T.reshape(-1)
    else:
        flulist = fc
        plist = data
    #print(flulist.shape, plist.shape)
    if masked > -360:
        plist = masked_where(plist <= masked, plist)
    return (flulist, plist)




# -*- coding: utf-8 -*-
#pylint: disable-msg=E0611, E1101, C0103, R0901, R0902, R0903, R0904, W0232
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------
"""Implements testing of environments and helpers.
"""

import unittest

from os.path import join

import numpy as np
#acoular imports
import acoular
acoular.config.global_caching = 'none' # to make sure that nothing is cached

from acoular import MicGeom, RectGrid3D, \
    SlotJet, OpenJet, RotatingFlow, \
    Environment, UniformFlowEnvironment, GeneralFlowEnvironment

# if this flag is set to True
WRITE_NEW_REFERENCE_DATA = False
# results are generated for comparison during testing.
# Should always be False. Only set to True if it is necessary to
# recalculate the data due to intended changes of the Beamformers.

m = MicGeom()
m.mpos_tot = ((0.5,0.5,0),(0,0,0),(-0.5,-0.5,0))
mc = m.mpos
g = RectGrid3D(x_min=-0.2, x_max=0.2, y_min=-0.2, y_max=0.2, z_min=0.5, z_max=0.9,
             increment=0.2 )
gc = g.gpos

flows = [
    SlotJet(v0=70.0, origin=(-0.7,0,0.7)),
    OpenJet(v0=70.0, origin=(-0.7,0,0.7)),
    RotatingFlow(v0=70.0, rpm=1000.0)
]

envs = [
    Environment(),
    UniformFlowEnvironment(ma=0.3),
    GeneralFlowEnvironment(ff=OpenJet(v0=70.0, origin=(-0.7,0,0.7)))
]


class acoular_env_test(unittest.TestCase):

    def test_flow_results(self):
        for fl in flows:
            with self.subTest(fl.__class__.__name__):
                name = join('reference_data',f'{fl.__class__.__name__}.npy')
                # stack all results
                actual_data = np.array([np.vstack(fl.v(x)) for x in gc.T])
                if WRITE_NEW_REFERENCE_DATA:
                    np.save(name,actual_data)
                ref_data = np.load(name)
                np.testing.assert_allclose(actual_data, ref_data, rtol=1e-5, atol=1e-8)

    def test_env_results(self):
        for env in envs:
            with self.subTest(env.__class__.__name__):
                name = join('reference_data',f'{env.__class__.__name__}.npy')
                # stack all results
                actual_data = np.vstack((env._r(gc,mc).T,env._r(gc).T))
                if WRITE_NEW_REFERENCE_DATA:
                    np.save(name,actual_data)
                ref_data = np.load(name)
                np.testing.assert_allclose(actual_data, ref_data, rtol=1e-5, atol=1e-8)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Example "Airfoil in open jet -- Beamforming".

Demonstrates barspectrum tool of Acoular.

A noise source emitting white noise is simulated. Its spectrum at a single 
microphone is plotted by using functions from acoular.tools.

Copyright (c)  Acoular Development Team.
All rights reserved.
"""

from acoular import WNoiseGenerator, PointSource, PowerSpectra, MicGeom, L_p
from acoular.tools import barspectrum
from numpy import array
from pylab import figure,plot,show,xlim,ylim,xscale,xticks,xlabel,ylabel,\
    grid,real, title, legend

# constants
sfreq= 12800 # sample frequency
band = 3 # octave: 1 ;   1/3-octave: 3 (for plotting)


# set up microphone at (0,0,0)
m = MicGeom()
m.mpos_tot = array([[0,0,0]])


# create noise source
n1 = WNoiseGenerator(sample_freq = sfreq, 
                     numsamples = 10*sfreq, 
                     seed = 1)

t = PointSource(signal = n1, 
                mics = m,  
                loc = (1, 0, 1))


# create power spectrum
f = PowerSpectra(time_data = t, 
                 window = 'Hanning', 
                 overlap = '50%', 
                 block_size = 4096)

# get spectrum data
spectrum_data = real(f.csm[:,0,0]) # get power spectrum from cross-spectral matrix
freqs = f.fftfreq() # FFT frequencies


# use barspectrum from acoular.tools to create third octave plot data
(f_borders, p, f_center) = barspectrum(spectrum_data, freqs, band, bar=True)
(f_borders_, p_, f_center_) = barspectrum(spectrum_data, freqs, band,bar=False)


# create figure with barspectra
figure(figsize=(20, 6))
title("Powerspectrum")
plot(f_borders,L_p(p), label="bar=True")
plot(f_borders_,L_p(p_),label="bar=False")
xlim(f_borders[0]*2**(-1./6),f_borders[-1]*2**(1./6))
ylim(50,90)
xscale('symlog')
label_freqs = [str(int(_)) for _ in f_center] # create string labels
xticks(f_center,label_freqs)
xlabel('f in Hz')
ylabel('SPL in dB')
grid(True)
legend()
show()



import nox

extras = ('', '[dev]', '[full]')

@nox.session
@nox.parametrize('extra', extras)
def build(session, extra):
    """Checks build for all supported Python versions and extras and optionally runs tests."""
    session.install(f'.{extra}')
    if extra == '[full]':
        with session.chdir('acoular/tests'):
            session.run('pip', 'list')
            session.run('/bin/bash', './run_tests.sh', external=True)



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------

from numpy import linspace, float64, float32
import acoular

# assume 64 microphones and 2048 grid points
# provide both 32 and 64 bit input data

g32 = linspace(1, 2, 3*2048, dtype=float32).reshape((3,2048)) 
m32 = linspace(2, 3, 192, dtype=float32).reshape((3,64))
g64 = g32.astype(float64)
m64 = m32.astype(float64)

def test_dist_mat32():
    acoular.environments.dist_mat(g32,m32)

def test_dist_mat64():
    acoular.environments.dist_mat(g64,m64)




from os import path
import numpy as np
import acoular
import unittest

rng = np.random.RandomState(2)
# test array
#np.abs(np.imag(ps.csm)).sum() + np.real(ps.csm).sum()
csm_sum = 0.5613882842200368
fft_sum = 0.3690552444079589+0.05264645224899272j

mg = acoular.MicGeom( mpos_tot=rng.normal(0,1,3*2).reshape(3,2))
sig = acoular.WNoiseGenerator(seed=1,numsamples=1010,sample_freq=1000)
p = acoular.PointSource(signal=sig,loc=(0,0,0),mics=mg)
ps = acoular.PowerSpectra( source=p, block_size=128, window='Hanning',cached=False )
fft = acoular.FFTSpectra(source=p,window='Hanning',block_size=128)

class Test_PowerSpectra(unittest.TestCase):

    def test_calc_csm(self):
        """ test that csm result has not changed over different releases"""
        test_csm_sum = np.abs(np.imag(ps.csm)).sum() + np.real(ps.csm).sum()
        self.assertAlmostEqual(test_csm_sum,csm_sum)

class Test_FFTSpectra(unittest.TestCase):

    def test_calc_fft(self):
        """ test that fft result has not changed over different releases"""
        test_fft_sum = 0
        for temp in fft.result():
            test_fft_sum += temp.sum()
        self.assertAlmostEqual(test_fft_sum,fft_sum)


if __name__ == '__main__':
    unittest.main()




# coding=UTF-8
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------

"""
The Acoular library: several classes for the implemetation of 
acoustic beamforming
"""

from .version import __author__, __date__, __version__

import os 

from .configuration import config

from .fileimport import time_data_import, csv_import, td_import, \
bk_mat_import, datx_import
try:
    from .nidaqimport import nidaq_import
except:
    pass

try:
    from .sdinput import SoundDeviceSamplesGenerator
except:
    pass

from .tbeamform import IntegratorSectorTime, \
BeamformerTime, BeamformerTimeSq, BeamformerTimeTraj, BeamformerTimeSqTraj,\
BeamformerCleant, BeamformerCleantSq, BeamformerCleantTraj, BeamformerCleantSqTraj
from .tprocess import SamplesGenerator, TimeInOut, MaskedTimeInOut, ChannelMixer, \
Mixer, TimeAverage, TimeReverse, TimePower, FiltFiltOctave, FiltOctave, TimeCache, \
WriteWAV, WriteH5, SpatialInterpolator, SpatialInterpolatorRotation, Trigger, \
SampleSplitter, AngleTracker, SpatialInterpolatorConstantRotation, Filter, \
TimeExpAverage, FiltFreqWeight, TimeCumAverage, FilterBank, OctaveFilterBank, TimeConvolve
from .calib import Calib
from .trajectory import Trajectory
from .grids import Grid, RectGrid, RectGrid3D, Sector, RectSector, RectSector3D, CircSector,\
    PolySector, MultiSector, MergeGrid, LineGrid, ImportGrid, ConvexSector
from .environments import cartToCyl, cylToCart, Environment, UniformFlowEnvironment, RotatingFlow, \
FlowField, OpenJet, SlotJet, GeneralFlowEnvironment
from .microphones import MicGeom
from .spectra import BaseSpectra, FFTSpectra, PowerSpectra, PowerSpectraImport, PowerSpectra as EigSpectra, synthetic

from .fbeamform import BeamformerBase, BeamformerCapon, BeamformerEig, \
BeamformerMusic, BeamformerDamas, BeamformerDamasPlus, BeamformerOrth, BeamformerCleansc, \
BeamformerCMF,BeamformerSODIX, BeamformerClean, BeamformerFunctional, BeamformerGIB, L_p, integrate, \
PointSpreadFunction, SteeringVector, BeamformerAdaptiveGrid, BeamformerGridlessOrth

from .sources import PointSource, MovingPointSource, \
TimeSamples, MaskedTimeSamples, PointSourceDipole, UncorrelatedNoiseSource, \
SourceMixer, SphericalHarmonicSource, LineSource, MovingPointSourceDipole, \
MovingLineSource, PointSourceConvolve
from .signals import SineGenerator, WNoiseGenerator, SignalGenerator,\
PNoiseGenerator, GenericSignalGenerator, FiltWNoiseGenerator

from . import tools

from . import demo



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
"""
import time as tm
import gc

import numpy as np
from numba import njit, guvectorize, complex128, void, prange

import sharedFunctions as shFncs
from beamformer import faverage  # The benchmark (created with scipy.weave)
from beamformer_withoutMP import faverage_OhneMP

#%% Numba - njit
@njit(complex128[:,:,:](complex128[:,:,:], complex128[:,:]))
def loops_Njit(csm, SpecAllChn):
    nFreqs = csm.shape[0]
    nMics = csm.shape[1]
    for cntFreq in range(nFreqs):
        for cntRow in range(nMics):
            temp = np.conj(SpecAllChn[cntFreq, cntRow])
            for cntColumn in range(nMics):
                csm[cntFreq, cntRow, cntColumn] += temp * SpecAllChn[cntFreq, cntColumn]
    return csm

@njit(complex128[:,:,:](complex128[:,:,:], complex128[:,:]), parallel=True)
def loops_Njit_Parallel(csm, SpecAllChn):
    nFreqs = csm.shape[0]
    nMics = csm.shape[1]
    for cntFreq in range(nFreqs):
        for cntRow in range(nMics):
            temp = np.conj(SpecAllChn[cntFreq, cntRow])
            for cntColumn in range(nMics):
                csm[cntFreq, cntRow, cntColumn] += temp * SpecAllChn[cntFreq, cntColumn]
    return csm

@njit(complex128[:,:,:](complex128[:,:,:], complex128[:,:]), parallel=True)
def loops_Njit_Parallel_Prange(csm, SpecAllChn):
    nFreqs = csm.shape[0]
    nMics = csm.shape[1]
    for cntFreq in range(nFreqs):
        for cntRow in prange(nMics):
            temp = np.conj(SpecAllChn[cntFreq, cntRow])
            for cntColumn in range(nMics):
                csm[cntFreq, cntRow, cntColumn] += temp * SpecAllChn[cntFreq, cntColumn]
    return csm

#%% create CSM via complex transpose of lower triangular matrix
@njit(complex128[:,:,:](complex128[:,:,:], complex128[:,:]))
def loopsComplexTranspose_Numpy(csm, SpecAllChn):
    nFreqs = csm.shape[0]
    for cntFreq in range(nFreqs):
        csm[cntFreq, :, :] += np.outer(np.conj(SpecAllChn[cntFreq, :]), SpecAllChn[cntFreq, :])
    return csm

@njit(complex128[:,:,:](complex128[:,:,:], complex128[:,:]))
def loopsOnlyTriangularMatrix_Njit(csm, SpecAllChn):
    """ one could only build the lower triangular csm and then, after averaging
    over all the ensenbles creating the whole csm by complex transposing.
    One could maybe use sparse CSR/CSC matrices (even though the CSM is not too big, so advantages are maybe small.)
    """
    nFreqs = csm.shape[0]
    nMics = csm.shape[1]
    for cntFreq in range(nFreqs):
        for cntRow in range(nMics):
            temp = np.conj(SpecAllChn[cntFreq, cntRow])
            for cntColumn in range(cntRow):  # only half of the operations in respect to 'loops_Njit'
                csm[cntFreq, cntRow, cntColumn] += temp * SpecAllChn[cntFreq, cntColumn]
    return csm

#%% Numba - guvectorize

# =============================================================================
# I don't think that parallelizing over the mics is in this case feasible.
# At least i can't think of a way to abstract the faverage procedure on one level below.
# It is however feasible to parallelize over the frequencies
# =============================================================================

@guvectorize([void(complex128[:,:], complex128[:], complex128[:,:])], '(m,m),(m)->(m,m)',
              nopython=True, target='cpu')
def loops_GuvectorizeOverFreqs_singleThreadedCPU(csm, SpecAllChn, result):
    nMics = csm.shape[0]
    for cntRow in range(nMics):
        temp = np.conj(SpecAllChn[cntRow])
        for cntColumn in range(nMics):
            result[cntRow, cntColumn] = csm[cntRow, cntColumn] + temp * SpecAllChn[cntColumn]

@guvectorize([void(complex128[:,:], complex128[:], complex128[:,:])], '(m,m),(m)->(m,m)',
              nopython=True, target='parallel')
def loops_GuvectorizeOverFreqs_multiThreadedCPU(csm, SpecAllChn, result):
    nMics = csm.shape[0]
    for cntRow in range(nMics):
        temp = np.conj(SpecAllChn[cntRow])
        for cntColumn in range(nMics):
            result[cntRow, cntColumn] = csm[cntRow, cntColumn] + temp * SpecAllChn[cntColumn]

@guvectorize([void(complex128[:,:], complex128[:], complex128[:,:])], '(m,m),(m)->(m,m)',
              nopython=True, target='cpu')
def loopsOnlyTriangularMatrix_GuvectorizeOverFreqs_singleThreadedCPU(csm, SpecAllChn, result):
    nMics = csm.shape[0]
    for cntRow in range(nMics):
        temp = np.conj(SpecAllChn[cntRow])
        for cntColumn in range(cntRow):
            result[cntRow, cntColumn] = csm[cntRow, cntColumn] + temp * SpecAllChn[cntColumn]

@guvectorize([void(complex128[:,:], complex128[:], complex128[:,:])], '(m,m),(m)->(m,m)',
              nopython=True, target='parallel')
def loopsOnlyTriangularMatrix_GuvectorizeOverFreqs_multiThreadedCPU(csm, SpecAllChn, result):
    nMics = csm.shape[0]
    for cntRow in range(nMics):
        temp = np.conj(SpecAllChn[cntRow])
        for cntColumn in range(cntRow):
            result[cntRow, cntColumn] = csm[cntRow, cntColumn] + temp * SpecAllChn[cntColumn]

#%%  MAIN
listOfMics = [500, 700, 1000]  # default: 64
listOfNFreqs = [2**cnt for cnt in range(4, 11)]  # default: 2048
nTrials = 10


#==============================================================================
# The benchmark function 'faverage' and also other implementations of
# the beamformer create a lot of overhead, which influences the computational
# effort of the succeding function. This is mostly the case, if concurrent
# calculations are done (multiple cores). So often the first trial of a new
# function takes some time longer than the other trials.
#==============================================================================

#funcsToTrial = [loopsComplexTranspose_Numpy, loops_Njit, loops_Njit_Parallel, loops_Njit_Parallel_Prange,
#                loopsOnlyTriangularMatrix_Njit, loops_GuvectorizeOverFreqs_singleThreadedCPU,
#                loops_GuvectorizeOverFreqs_multiThreadedCPU,
#                loopsOnlyTriangularMatrix_GuvectorizeOverFreqs_singleThreadedCPU,
#                loopsOnlyTriangularMatrix_GuvectorizeOverFreqs_multiThreadedCPU,
#                faverage_OhneMP, faverage]
funcsToTrial = [loopsOnlyTriangularMatrix_Njit, faverage]

for nMics in listOfMics:
    for nFreqs in listOfNFreqs:
        # Init
        print(10*'-' + 'New Test configuration: nMics=%s, nFreqs=%s' %(nMics, nFreqs) + 10*'-')
        print(10*'-' + 'Creation of inputInputs' + 10*'-')

        csm = np.zeros((nFreqs, nMics, nMics), np.complex128)
        spectrumInput = np.random.rand(nFreqs, nMics) + \
                1j*np.random.rand(nFreqs, nMics)

        nameOfFuncsToTrial = map(lambda x: x.__name__, funcsToTrial)
        nameOfFuncsForError = [funcName for funcName in nameOfFuncsToTrial if funcName != 'faverage']
        maxRelativeDeviation = np.zeros((len(funcsToTrial), nTrials))
        maxAbsoluteDeviation = np.zeros((len(funcsToTrial), nTrials))
        timeConsumption = [[] for _ in range(len(funcsToTrial))]
        indOfBaselineFnc = nameOfFuncsToTrial.index('faverage')

        print(10*'-' + 'Onetime calculation of "faverage" for error reference' + 10*'-')
        faverage(csm, spectrumInput)
        resultReference = csm  # For relative/absolute error
        gc.collect()

        # Testing
        print(10*'-' + 'Testing of functions' + 10*'-')
        cntFunc = 0
        for func in funcsToTrial:
            print(func.__name__)
            for cntTrials in xrange(nTrials):
                csm = np.zeros((nFreqs, nMics, nMics), np.complex128)
                resultHelp = np.zeros((nFreqs, nMics, nMics), np.complex128)
                if func.__name__ == 'faverage' or func.__name__ == 'faverage_OhneMP':
                    t0 = tm.time()
                    func(csm, spectrumInput)
                    t1 = tm.time()
                    result = csm
                elif func.__name__ == 'loops_GuvectorizeOverFreqs':
                    t0 = tm.time()
                    func(csm, spectrumInput, resultHelp)
                    t1 = tm.time()
                    result = resultHelp
                else:
                    t0 = tm.time()
                    output = func(csm, spectrumInput)
                    t1 = tm.time()
                    result = output
                timeConsumption[cntFunc].append(t1 - t0)
                relativeDiffBetweenNewCodeAndRef = (result - resultReference) / (result + resultReference) * 2  # error in relation to the resulting value
                maxRelativeDeviation[cntFunc, cntTrials] = np.amax(np.amax(np.amax(abs(relativeDiffBetweenNewCodeAndRef), axis=0), axis=0), axis=0) + 10.0**-20  # relative error in inf-norm
                maxAbsoluteDeviation[cntFunc, cntTrials] = np.amax(np.amax(np.amax(abs(result - resultReference), axis=0), axis=0), axis=0) + 10.0**-20  # absolute error in inf-norm
            cntFunc += 1
        factorTimeConsump = [np.mean(timeConsumption[cnt]) for cnt in range(0, len(funcsToTrial))] \
                            / np.mean(timeConsumption[indOfBaselineFnc])

        # Save the current test-config as .sav
        helpString = 'The order of the variables is: \n nameOfFuncsToTrial \n maxRelativeDeviation'\
                '\n timeConsumption [nFuncs, nTrials] \n nMics \n nGridPoints \n nFreqs '\
                '\n Factor of time consumption (in relation to the original .cpp) \n maxAbsoluteDeviation \n nThreadsGlobal'
        saveTupel = (helpString, nameOfFuncsToTrial, maxRelativeDeviation, timeConsumption,
                     nMics, 0, nFreqs, factorTimeConsump, maxAbsoluteDeviation, 0)
        stringParameters = 'faverage_TestcasesTimeConsumption_nMics%s_nFreqs%s_nTrials%s' %(nMics, nFreqs, nTrials)

#        stringSaveName = 'Peter'
        stringSaveName = 'Sicherung_DurchgelaufeneTests/faverage/' + stringParameters

        shFncs.savingTimeConsumption(stringSaveName, saveTupel)  # saving as "stringSaveName.sav"
#        shFncs.plottingOfOvernightTestcasesBeamformer(stringSaveName + '.sav')  # plot of the current test-config




# -*- coding: utf-8 -*-
#pylint: disable-msg=E0611, E1101, C0103, R0901, R0902, R0903, R0904, W0232
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------
"""Implements testing of frequency beamformers.
"""

import unittest

from os.path import join

import numpy as np

#acoular imports
import acoular
acoular.config.global_caching = 'none' # to make sure that nothing is cached

from acoular import Calib, MicGeom, PowerSpectra, \
RectGrid, BeamformerBase, BeamformerEig, BeamformerOrth, BeamformerCleansc, \
MaskedTimeSamples, BeamformerCMF, \
BeamformerCapon, BeamformerMusic, BeamformerDamas, BeamformerClean, \
BeamformerFunctional, BeamformerDamasPlus, BeamformerGIB, BeamformerGridlessOrth,\
SteeringVector, Environment

# if this flag is set to True
WRITE_NEW_REFERENCE_DATA = False
# new beamformer results are generated for comparison during testing. 
# Should always be False. Only set to True if it is necessary to 
# recalculate the data due to intended changes of the Beamformers.


#load exampledata
datafile = join('..','..','examples','example_data.h5')
calibfile = join('..','..','examples','example_calib.xml')
micgeofile = join( '..','xml','array_56.xml')

#frequencies to test
cfreqs = 1000,8000

# values from example 1
t1 = MaskedTimeSamples(name=datafile)
t1.start = 0 # first sample, default
t1.stop = 16000 # last valid sample = 15999
invalid = [1,7] # list of invalid channels (unwanted microphones etc.)
t1.invalid_channels = invalid 
t1.calib = Calib(from_file=calibfile)
m = MicGeom(from_file=micgeofile)
m.invalid_channels = invalid
g = RectGrid(x_min=-0.6, x_max=-0.0, y_min=-0.3, y_max=0.3, z=0.68,
             increment=0.05 )
env=Environment(c=346.04)
st = SteeringVector(grid=g, mics=m, env=env)
f = PowerSpectra(time_data=t1, 
               window='Hanning', overlap='50%', block_size=128, #FFT-parameters
               cached = False )  

# produces a tuple of beamformer objects to test
# because we need new objects for each test we have to call this more than once
def fbeamformers():
    bb = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = False)
    be = BeamformerEig(freq_data=f, steer=st, r_diag=True, n=54, cached = False)

    #frequency beamformers to test
    bbase = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = False)
    bc = BeamformerCapon(freq_data=f, steer=st, cached=False)
    beig = BeamformerEig(freq_data=f, steer=st, r_diag=True, n=54, cached = False)
    bm = BeamformerMusic(freq_data=f, steer=st, n=6, cached = False)
    bd = BeamformerDamas(beamformer=bb, n_iter=10, cached = False)
    bdp = BeamformerDamasPlus(beamformer=bb, n_iter=100, cached = False)
    bo = BeamformerOrth(freq_data=f, steer=st, r_diag=True, eva_list=list(range(38,54)), cached = False)
    bs = BeamformerCleansc(freq_data=f, steer=st, r_diag=True, cached = False)
    bcmf = BeamformerCMF(freq_data=f, steer=st, method='LassoLarsBIC', cached = False)
    bl = BeamformerClean(beamformer=bb, n_iter=10, cached = False)
    bf = BeamformerFunctional(freq_data=f, steer=st, r_diag=False, gamma=3, cached = False)
    bgib = BeamformerGIB(freq_data=f, steer=st, method= 'LassoLars', n=2, cached = False)
    bgo = BeamformerGridlessOrth(freq_data=f, steer=st, r_diag=False, n=1,  shgo={'n':16}, cached = False)
    return (bbase, bc, beig, bm, bl, bo, bs, bd, bcmf, bf, bdp, bgib, bgo)

class acoular_beamformer_test(unittest.TestCase):

    def test_beamformer_freq_results(self):
        # we expect the results to computed
        acoular.config.global_caching = 'none'
        for b in fbeamformers():
            with self.subTest(b.__class__.__name__+" global_caching = none"):
                name = join('reference_data',f'{b.__class__.__name__}.npy')
                # stack all frequency band results together
                actual_data = np.array([b.synthetic(cf,1) for cf in cfreqs],dtype=np.float32)
                if WRITE_NEW_REFERENCE_DATA:
                    np.save(name,actual_data)
                ref_data = np.load(name)
                np.testing.assert_allclose(actual_data, ref_data, rtol=5e-5, atol=5e-8)
        # we expect the results to be computed and written to cache
        acoular.config.global_caching = 'individual'
        for b in fbeamformers():
            b.cached = True
            with self.subTest(b.__class__.__name__+" global_caching = individual"):
                name = join('reference_data',f'{b.__class__.__name__}.npy')
                actual_data = np.array([b.synthetic(cf,1) for cf in cfreqs],dtype=np.float32)
                ref_data = np.load(name)
                np.testing.assert_allclose(actual_data, ref_data, rtol=5e-5, atol=5e-8)
        # we expect the results to be read from cache
        acoular.config.global_caching = 'all'
        for b in fbeamformers():
            b.cached = True
            with self.subTest(b.__class__.__name__+" global_caching = all"):
                name = join('reference_data',f'{b.__class__.__name__}.npy')
                actual_data = np.array([b.synthetic(cf,1) for cf in cfreqs],dtype=np.float32)
                ref_data = np.load(name)
                np.testing.assert_allclose(actual_data, ref_data, rtol=5e-5, atol=5e-8)
        # we expect the cached results to be overwritten
        acoular.config.global_caching = 'overwrite'
        for b0,b1 in zip(fbeamformers(),fbeamformers()):
            b0.cached = True
            b1.cached = True
            with self.subTest(b0.__class__.__name__+" global_caching = overwrite"):
                if hasattr(b0,'beamformer'): # BeamformerClean, BeamformerDamas, BeamformerDamasplus 
                                            # do not pass because the .beamformer result is not take from cache
                    continue                         # nor recalculated   
                b0.result[:] = 0
                self.assertFalse(np.any(b0.result))
                name = join('reference_data',f'{b1.__class__.__name__}.npy')
                actual_data = np.array([b1.synthetic(cf,1) for cf in cfreqs],dtype=np.float32)
                ref_data = np.load(name)
                np.testing.assert_allclose(actual_data, ref_data, rtol=5e-5, atol=5e-8)

    def test_beamformer_caching(self):
        # within each subcase, we need new beamformer objects because result is not updated when
        # global_caching or cached changes
        with self.subTest("global_caching = 'none'"):
            acoular.config.global_caching = 'none'
            b0 = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = True)
            b1 = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = True)
            self.assertNotEqual(id(b0.result),id(b1.result))
        with self.subTest("global_caching = 'individual'"):
            acoular.config.global_caching = 'individual'
            b0 = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = True)
            b1 = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = True)
            self.assertEqual(id(b0.result),id(b1.result))
            b1 = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = False)
            self.assertNotEqual(id(b0.result),id(b1.result))
        with self.subTest("global_caching = 'all'"):
            acoular.config.global_caching = 'all'
            b0 = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = True)
            b1 = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = False)
            self.assertEqual(id(b0.result),id(b1.result))
        with self.subTest("global_caching = 'readonly'"):
            acoular.config.global_caching = 'readonly'
            b0 = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = True)
            b1 = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = True)
            self.assertEqual(id(b0.result),id(b1.result))

class Test_PowerSpectra(unittest.TestCase):

    def test_csm(self):
        """ test that csm result has not changed over different releases"""
        name = join('reference_data',f'{f.__class__.__name__}_csm.npy')
        # test only two frequencies
        actual_data = np.array(f.csm[(16,32),:,:],dtype=np.complex64)
        if WRITE_NEW_REFERENCE_DATA:
            np.save(name,actual_data)
        ref_data = np.load(name)
        np.testing.assert_allclose(actual_data, ref_data, rtol=1e-5, atol=1e-8)

    def test_ev(self):
        """ test that eve and eva result has not changed over different releases"""
        name = join('reference_data',f'{f.__class__.__name__}_ev.npy')
        # test only two frequencies
        actual_data = np.array((f.eve*f.eva[:,:,np.newaxis])[(16,32),:,:],dtype=np.complex64)
        if WRITE_NEW_REFERENCE_DATA:
            np.save(name,actual_data)
        ref_data = np.load(name)
        np.testing.assert_allclose(actual_data, ref_data, rtol=1e-5, atol=1e-8)


class TestSteerFormulation(unittest.TestCase):

    def test_all_steer_formulation(self):
        """ tests all variants of beamformerFreq subroutines
        """
        st = SteeringVector(grid=g, mics=m, env=env)
        b0 = BeamformerBase(freq_data=f, steer=st, cached = False)
        b1 = BeamformerEig(freq_data=f, steer=st, n=54, cached = False)
        for ki,kind in enumerate(('classic', 'inverse','true level', 'true location')):
            st.steer_type = kind
            for b in (b0,b1):
                for dr in (True,False):
                    b.r_diag = dr
                    with self.subTest(f"{b.__class__.__name__} r_diag:{dr} steer:{kind}"):
                        name = join('reference_data',f"{b.__class__.__name__}{dr}{ki+1}.npy")
                        actual_data = np.array([b.synthetic(cf,1) for cf in cfreqs],dtype=np.float32)
                        if WRITE_NEW_REFERENCE_DATA:
                            np.save(name,actual_data)
                        ref_data = np.load(name)
                        np.testing.assert_allclose(actual_data, ref_data, rtol=1e-5, atol=1e-8)


if __name__ == '__main__':
    unittest.main() #exit=False







# -*- coding: utf-8 -*-
"""
Plot Acoular examples

Runs the example scripts found in the current directoy (example*.py)
and exports generated figures.
This is for easily creating the figures used for the documentation.

Usage: 
- Copy script into acoular/examples/ directory
- Run in terminal
- Move generated *.png files to acoular/docs/source/examples/

Do not run this script in an ipython console.

Copyright (c) 2018 The Acoular developers.
All rights reserved.
"""

from glob import glob
from importlib import import_module
from pylab import get_fignums, figure, savefig, close

for ex in glob("example*.py"):
    exname = ex[:-3]
    print('Importing %s ...' % exname)
    try:
        import_module(exname)
        for fn in get_fignums():
            figure(fn)
            figname = exname + '_' + str(fn)+ '.png'
            print('Exporting %s ...' % figname)
            savefig(figname, bbox_inches='tight')
        close('all')
    except:
        print('---------------------------------------------')
        print('        Error importing %s !' % ex)
        print('---------------------------------------------')
